
#include "stream.h"

struct stream
{
    stream_type_t type;

    buffer_t * buffer;
    buffer_t * cache;   /* The cache operates over the same data space as buffer to omit unnecessary copying data between two buffers.
                         * Instead, we synchronize the buffer and cache by updating their read and write positions.
                         */
};

ret_code_t _stream_init( stream_t * _stream, const stream_conf_t * _conf )
{
    ret_code_t result = NRF_ERROR_INTERNAL;
    buffer_conf_t buffer_conf;

    ASSERT_RULE( _conf->type == STREAM_TYPE_INPUT || _conf->type == STREAM_TYPE_OUTPUT, "Invalid stream type detected." );

    _stream->type = _conf->type;

    /* Initialize the stream buffer */
    result = buffer_init( &_stream->buffer, &_conf->buffer);
    EXIT_IF_ERR( result, "buffer_init (buffer) failed" );

    /* Initialize the stream cache - initially similar to the stream buffer */
    buffer_conf = buffer_get_conf( _stream->buffer );
    result = buffer_init( &_stream->cache, &buffer_conf );
    EXIT_IF_ERR( result, "buffer_init (cache) failed" );

_EXIT:
    return result;
}

ret_code_t stream_init( stream_t ** __stream, const stream_conf_t * _conf )
{
    ret_code_t result = NRF_ERROR_INTERNAL;

    /* Allocate the gpio instance */
    *__stream = heap_alloc( sizeof(stream_t) );

    result = _stream_init( *__stream, _conf );
    EXIT_IF_ERR( result, "_stream_init failed" );

_EXIT:
    return result;
}

buffer_t * stream_get_buffer( stream_t * _stream )
{
    return _stream->buffer;
}

buffer_t * stream_get_cache( stream_t * _stream )
{
    return _stream->cache;
}

void _input_cache_sync( stream_t * _stream )
{
    uint16_t buffer_read_pos;
    uint16_t cache_write_pos;

    /* Get the new read and write positions */
    buffer_read_pos = buffer_get_read_pos( _stream->buffer );
    cache_write_pos = buffer_get_write_pos( _stream->cache );

    /* Synchronize the cache with the RX buffer */
    buffer_set_read_pos_safe( _stream->cache, buffer_read_pos );
    buffer_set_write_pos_safe( _stream->buffer, cache_write_pos );
}

void _output_cache_sync( stream_t * _stream )
{
    uint16_t buffer_write_pos;
    uint16_t cache_read_pos;

    /* Get the new read and write positions */
    buffer_write_pos = buffer_get_write_pos( _stream->buffer );
    cache_read_pos = buffer_get_read_pos( _stream->cache );

    /* Synchronize the cache with the RX buffer */
    buffer_set_write_pos_safe( _stream->cache, buffer_write_pos );
    buffer_set_read_pos_safe( _stream->buffer, cache_read_pos );
}

/*
 * The cache synchronization should be run inside a protected space (e.g. a MUTEX constrained space) to prevent hazardous behavior.
 */
void stream_cache_sync( stream_t * _stream )
{
    switch( _stream->type )
    {
        case STREAM_TYPE_INPUT:

            _input_cache_sync( _stream );

            break;

        case STREAM_TYPE_OUTPUT:

            _output_cache_sync( _stream );

            break;
    }
}
